/*
 * StringBufferPool.java
 *
 * Created on January 17, 2003, 3:33 PM
 */

package com.divnull.util;

import java.util.Iterator;
import java.util.Set;
import org.apache.commons.pool.impl.StackKeyedObjectPool;

/** A stack-based keyed pool specialized to handle <code>StringBuffer</code>
 * objects.
 * <p>
 * If used through an entire application, a string buffer pool offers
 * surprisingly significant performance improvement to Java applications at the
 * cost of slightly more complicated code. (In particular, programmers must
 * remember to release the string buffers they use.)
 * <p>
 * This string buffer pool implementation uses the initial size of the buffer
 * as its key in the pool. Unlike most key-based pools, the programmer is not
 * responsible for passing this same key to the pool upon release, as the
 * string buffer knows its own size. Internally, since string buffers can grow
 * with use, a string buffer returned to the pool may be made available under
 * a different key than the one with which it was first allocated.
 * <p>
 * This pool offers a number of "standard" buffer sizes and helper routines to
 * use them. Using these standard sizes helps optimize performance, as it
 * avoid allocating temporary <code>Integer</code> objects. Performance is also
 * significantly improved when programmers make sure to ask for a buffer large
 * enough to avoid growth while they use it.
 * <p>
 * While this pool can be used on its own, most users will find it more
 * convenient to use the singleton class {@link SBPool}, which wraps the pool
 * in a single, globally accessible access point and sets up the proper factory
 * for use with the pool. Instructions for using this pool can be found in the
 * documentation of that class.
 *
 * @author  Wordman
 * @see SBPool
 * @see StringBufferPoolableObjectFactory
 */
public class StringBufferPool extends StackKeyedObjectPool
{
    public final static Integer kIgnoredKey = new Integer(-1);
    public final static Integer kMicroStringSize = new Integer(32);
    public final static Integer kTinyStringSize = new Integer(64);
    public final static Integer kSmallStringSize = new Integer(128);
    public final static Integer kMediumStringSize = new Integer(256);
    public final static Integer kLargeStringSize = new Integer(512);
    public final static Integer kHugeStringSize = new Integer(1024);
    public final static Integer kGiantStringSize = new Integer(2048);
    public final static Integer kGargantuanStringSize = new Integer(4096);
    
    /*------- Overridden methods -------*/
    
    /** The standard stack-based keyed object pool requires the user to
     * remember the key they used to borrow the object. With string buffers,
     * this is not neccessary and can actually be undesirable, because
     * the key is the size of the buffer, and this can grow over time.
     *
     * <p>This method overrides the default behavior of the stack-based keyed
     * object pool by ignoring the key value completely and looking at the
     * current size of the string. It is put back into the pool under a key
     * appropriate to its new size.
     *
     * <p>In general, no one should actually call this method directly. Instead
     * call {@link #returnBuffer}. In an ideal world, this method should be
     * made protected instead of public, but this would violate the interface.
     *
     * @param key This value is ignored.
     * @param obj The buffer to return to the pool.
     * @throws Exception Thrown during an error with the pool.
     * @see #returnBuffer
     * @see #returnBufferGetString
     */
    public void returnObject(Object key,Object obj) throws Exception
    {
        if ( obj == null || key == null )
            return;
        
        StringBuffer buffer = (StringBuffer)obj;
        
        /* When returning a string, its size may have changed from when it
         * was first allocated, so it will not necessarily return back
         * into the stack from where it first came. Find the best of the
         * current stacks to put it in.
         */
        int size = buffer.capacity();
        
        /* Check to see if the passed in key happens to match the size.
         * This can happen when people call this method directly instead
         * of using {@link #returnBuffer}.
         */
        Integer newKey = (Integer)key;
        if ( newKey.intValue() != size )
            newKey = findOptimalReturnKey(size);
        
        /* Need to make sure the string is big enough. Example, a user asks
         * for a 32 character buffer, but during use, it grows to 66 bytes.
         * It is then returned. The proper key for this buffer is 128, as it 
         * is the smallest standard key larger than the size of the buffer.
         * To return it to the 
        
        /* Now, return the buffer to the pool under the new key. */
        super.returnObject( newKey,obj );
    }
    
    
    /** In most keyed pools, this method would return the number of objects
     * currently active under this key. Since string buffers are keyed by 
     * size and, while active, the size can change, the count returned by the
     * method cannot be easily ascertained, and is somewhat meaningless.
     * <p>
     * The standard stack keyed pool count will sometimes be accurate and 
     * sometimes not. This pool elects to override this behavior and always
     * return -1 instead (generally it is better to provide something that is 
     * <em>always</em> inaccurate as opposed to <em>sometimes</em> inaccurate).
     * <p>
     * Not implementing this function is considered less irritating than
     * requiring programmers to return the correct key value when they return
     * the buffer.
     */
    public int getNumActive(Object key)
    {
        return -1;
    }
    
	/** Get the number of idle strings of the requested size. This is overridden
	 * to pin the requested size to one of the standard sizes.
	 * @see org.apache.commons.pool.KeyedObjectPool#getNumIdle(Object)
	 */
	public int getNumIdle(Object obj)
	{
		Integer key = (Integer)obj;
		key = findOptimalBorrowKey(key.intValue());
		return super.getNumIdle(key);
	}
    
    /*------- Implementation methods -------*/
    
    /** Prepopulate the pool with a given number of string buffers of a given
     * size.
     * @param size The size to allocate the string buffers. This will be pinned
     * to a standard size, if possible.
     * @param count The number of buffers of that size to create. */
    public void prePopulatePool( int size,int count ) throws Exception
    {
        Integer key = findOptimalBorrowKey(size);
        int index = count - getNumIdle(key);
        while ( index > 0 )
        {
            StringBuffer buffer = (StringBuffer)borrowObject(key);
            returnObject(key,buffer);
            --index;
        }
    }

    /** Retrieves the key value best suited to used for borrowing a buffer of 
     * the passed in size, creating a new key if needed.
     * <p>The method searches for an existing Integer object in use as a key
     * in the pool. If it finds one that matches the passed in size, it uses
     * it. If not, the concern is that the user be given a buffer <em>at 
     * least</em> as large as the one they asked for. So, for example, if
     * they ask for a 66 characters buffer, the pool will give them the smallest 
     * standard buffer size that still contains 66 characters (i.e. 128).
     * @param size The size of the buffer to look for.
     * @param forBorrow If getting a key for a borrow operation, 
     * @return An available key of the proper size. Should never be null. */
    protected Integer findOptimalBorrowKey( int requestedSize )
    {
        /* The findOptimalReturnKey method finds the opposite of what we want
         * here: the greatest key that is smaller than the size passed to it.
         * This will be exactly one standard size less than we want. So, we 
         * take advantage of the fact that our standard sizes are powers of 
         * two and double our size request (minus 1).
         */
        int size = (requestedSize * 2) - 1;
        Integer foundKey = findOptimalReturnKey(size);
        
        /* In unusual cases (such as very large buffers), the standard sizes
         * may not suffice. In this case, keys are created as needed.
         */
        if ( foundKey == null || foundKey.intValue() < requestedSize )
            foundKey = generateKey(requestedSize);
        return foundKey;
    }

    /** Retrieves the key value best suited to used for a buffer of the passed
     * in size, creating a new key if needed.
     * <p>The method searches for an existing Integer object in use as a key
     * in the pool. If it finds one that matches the passed in size, it uses
     * it. If not, the concern is that the buffer get returned to a stack 
     * with a key <em>at least</em> as large as the buffer, but not larger. This
     * will guarentee that if that same buffer is borrow later, it will be at
     * least as large as the key of that stack. So, for example, if a 66 
     * character buffer is returned, it should be placed into the 64 character
     * stack. Were it placed into the 128 character stack and someone asked for
     * a 128 character buffer, it might get returned when is is not really that
     * large! By putting it in the 64 character stack, it will only get returned
     * when someone asks for a 64 character buffer.
     * <p>Formally, this method will return the key returned is the greatest 
     * existing key that is smaller than the passed in size. If none exist 
     * (which should happen only when the size is larger than any standard key),
     * a new key is allocated of the passed in size.
     * @param size The size of the buffer being returned.
     * @return An available key of the proper size. Should never be null. */
    protected Integer findOptimalReturnKey( int size )
    {
        /* There is a bit of a performance conundrum here. Because size is
         * extremely likely to have the exact value of an existing key,
         * it is tempting to call _pools.containsKey. This would use fewer
         * function calls and is probably O(lg n) instead of O(n) (like
         * the code below), so would probably amortize to being faster
         * the the code below alone. The problem is that do do this, we'd
         * have to allocate an Integer object to do the check every single
         * time a string was returned to the pool. Since the whole point
         * of this pool is to avoid allocating tons a small objects, this
         * would be self defeating.
         *
         * One quick, though imperfect optimization is to see if any of the
         * standard keys match the string size. This should happen in the
         * majority of cases. Since these objects are already allocated,
         * they can be used to check against the size.
         */
        Integer foundKey = findStandardKey(size);
        
        Set stackSizes = _pools.keySet();
        if ( foundKey == null && stackSizes != null )
        {
            /* Find the largest of the current stacks with a size less than
             * the current buffer size. This search is needed, because it 
             * is possible that non-standard keys are in use for very large
             * buffers.
             */
            int bestKey = 0;
            int currentKey = 0;
            Iterator itr = stackSizes.iterator();
            while ( itr.hasNext() )
            {
                Integer current = (Integer)itr.next();
                currentKey = current.intValue();
                if (currentKey <= size && currentKey >= bestKey)
                {
                    bestKey = currentKey;
                    foundKey = current;
                }
            }
        }
        
        /* If a key could not be found, make a new one. This should happen only
         * if the requested size is huge.
         */
        if ( foundKey == null )
            foundKey = generateKey(size);
        
        return foundKey;
    }
    
    /** Retrieves the standard key value identifying the stack to which a 
     * buffer of a given size should be returned.
     * <p>The method finds the constant representing a standard buffer size
     * with the greatest value that is still smaller than the passed in size.
     * @param size The size of the stack to look for.
     * @return An available key of the proper standad size, or null if one does
     * not exist. */
    protected Integer findStandardKey(int size)
    {         
        /* Run a bastardized pseudo binary search. Done with if statements,
         * since all the standard ways of doing this in Java require allocation
         * of Integer objects, which we are trying to avoid.
         * Really, really ugly, but it works. I'm so proud (not).
         *
         * A slower, though more understandable, version that would return 
         * identical output would be:
         *      if ( size > kGargantuanStringSize.intValue() )
         *          return kGargantuanStringSize;
         *      if ( size > kGiantStringSize.intValue() )
         *          return kGiantStringSize;
         *      <...and so on down the sizes...>
         *      if ( size > kTinyStringSize.intValue() )
         *          return kTinyStringSize;
         *      else
         *          return kMicroStringSize;
         */
        if ( size < kLargeStringSize.intValue() )
        {
            if ( size < kSmallStringSize.intValue() )
            {
                if ( size >= kTinyStringSize.intValue() )
                    return kTinyStringSize; // Size between kTiny and kSmall-1
                
                /* Slight exception to the main rule, in that kMicroStringSize
                 * represents strings from 0 to kTinyStringSize-1, rather than
                 * kMicroStringSize to kTinyStringSize-1.
                 */
                return kMicroStringSize;  // Size between 0 and kTiny-1
            }
            else
            {
                if ( size >= kMediumStringSize.intValue() )
                    return kMediumStringSize; // Size between kMedium and kLarge-1
                return kSmallStringSize; // Size between kSmall and kMedium-1
            }
        }
        else
        {
            if ( size < kGiantStringSize.intValue() )
            {
                if ( size >= kHugeStringSize.intValue() )
                    return kHugeStringSize; // Size between kHuge and kGiant-1
                return kLargeStringSize; // Size between kLarger and kHuge-1
            }
            else
            {
                if ( size <= kGargantuanStringSize.intValue() )
                    return kGiantStringSize; // Size between kGiant and kGargantuan-1
                if ( size == kGargantuanStringSize.intValue() )
                    return kGargantuanStringSize; // Size exactly kGargantuan
            }
        }
        
        /* If the size is larger than kGargantuanStringSize, don't use a standard
         * key. This should happen very rarely, as most people will use streams
         * for buffers larger than 4k.
         */
        return null;
    }
    
    /** Generates a new, non-standard key. This is called
     * if a standard key cannot be found and a key of the requested size does
     * not already exist. This implementation simply generates a key exactly
     * equal in value to the requested size. This exists as a separate 
     * method largely so it can be overridden.
     * @param size The size of the buffer for which to generate a key.
     * @return A key appropriate to the passed in size.
     */
    protected Integer generateKey(int size)
    {
        Integer key = new Integer(size);
        return key;
    }
    
    
    
    /*------- Utility methods -------*/
    
    /* NOTE: Most of these methods are declared final so that JIT compilers
     * can optimize calls to them by inlining. This may provide little
     * benefit, however. Still, it is not likely that these methods will
     * be overridden.
     */
    
    /* NOTE: Normally, it is bad form to use exact byte counts in function
     * names. In this case, however, it proved helpful. At first, I had names
     * like <code>borrowMicroBuffer</code> and <code>borrowLargeBuffer</code>,
     * but programmers were constantly needing to look up how big each of these
     * allocations were. The information they had in hand was how many
     * characters they wanted, so using methods like <code>borrowBuffer32</code>
     * and <code>borrowuffer512</code> sped up development.
     */
    
    /** Get a string buffer preallocated to an arbitrary size.
     * <p>
     * This routine should be avoided, when possible, in favor of one of the
     * methods that return a buffer of one of the "standard" sizes. Performance
     * of the pool is optimized when these standard sizes are used.
     * @param size The preallocated size of the buffer.
     * @return A buffer of the requested size.
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */
    public StringBuffer borrowBuffer( int size ) throws Exception
    {
        /* Check to see if the requested size matches one of our standard
         * sizes. This will prevent an Integer object allocation.
         */
        Integer bufSize = findOptimalBorrowKey(size);
        StringBuffer buffer = (StringBuffer)borrowObject(bufSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 32 characters.
     * @return A 32 character buffer */
    public final StringBuffer borrowBuffer32() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kMicroStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 64 characters.
     * @return A 64 character buffer */
    public final StringBuffer borrowBuffer64() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kTinyStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 128 characters.
     * @return A 128 character buffer */
    public final StringBuffer borrowBuffer128() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kSmallStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 256 characters.
     * @return A 256 character buffer */
    public final StringBuffer borrowBuffer256() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kMediumStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 512 characters.
     * @return A 512 character buffer */
    public final StringBuffer borrowBuffer512() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kLargeStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 1024 characters.
     * @return A 1024 character buffer */
    public final StringBuffer borrowBuffer1k() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kHugeStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 2048 characters.
     * @return A 2048 character buffer */
    public final StringBuffer borrowBuffer2k() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kGiantStringSize);
        return buffer;
    }
    
    /** Get a string buffer preallocated to 4096 characters.
     * @return A 4096 character buffer */
    public final StringBuffer borrowBuffer4k() throws Exception
    {
        StringBuffer buffer = (StringBuffer)borrowObject(kGargantuanStringSize);
        return buffer;
    }
    
    /** Return a string buffer to the pool.
     * @param buffer A buffer previously borrowed from the pool.
     * @see #returnBufferGetString
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */
    public void returnBuffer(StringBuffer buffer) throws Exception
    {
        returnObject( kIgnoredKey,buffer );
    }
    
    /** Return a string buffer to the pool and get back the string object
     * from the buffer.
     * @param buffer A buffer previously borrowed from the pool.
     * @return The string object rendered from the buffer.
     * @see #returnBuffer
     * @see #borrowBuffer32
     * @see #borrowBuffer64
     * @see #borrowBuffer128
     * @see #borrowBuffer256
     * @see #borrowBuffer512
     * @see #borrowBuffer1k
     * @see #borrowBuffer2k
     * @see #borrowBuffer4k
     */
    public String returnBufferGetString(StringBuffer buffer) throws Exception
    {
        String data = buffer.toString();
        returnBuffer( buffer );
        return data;
    }
}
